#block.intern and block.extern can also have spyderclass None to have only trigger I/O
##"redirect" specifies a path to find out the *spyderclass* later: the *data* is always intern or extern!
#intern: push antenna, pull output
#extern: pull antenna, pull output


import libcontext
from libcontext.pluginclasses import *
from libcontext.socketclasses import *
import functools

from .event import exception
from .segments._runtime_segment import tryfunc

import spyder, Spyder

class general(object): pass

class dummy(object): 
  def __getattr__(self, attr):
    ret = dummy()
    setattr(self, attr, ret)
    return ret


from . import drone

class blockbase(drone):
  def add_exc_target(self, target):  
    self._exc_targets.append(target)

  def _add_trig_target(self, target):
    red = self._get_redirect()
    if red: return red._add_trig_target(target)  
    self._trig_targets.append(target)

  def _trig(self):    
    red = self._get_redirect()
    if red: return red._trig()  
    for t in self._trig_targets: t()

  def _get_redirect(self):
    if self._redirect:
      if isinstance(self._redirect, str):
        redirect = getattr(self.parent, self._redirect)
      else:
        redirect = self.parent        
        for attr in self._redirect:
          if attr.startswith("$"):
            redirect = [r[1] for r in redirect if r[0] == attr[1:]][0]            
          else:
            redirect = getattr(redirect, attr)
      self._spyderclass = redirect._spyderclass
      return redirect
    
  def catch(self, segmentname, exc_type, exc_value):
    if segmentname is None:
      e = exception((self.parent.hivename,self.beename), (exc_type, exc_value))
    else:
      e = exception((self.parent.hivename,self.beename,segmentname), (exc_type, exc_value))
    for tnr,t in enumerate(self._exc_targets):
      t(e)
      if e.cleared: break
    return e
  def place(self): pass
  
class intern(blockbase):
  def __init__(self, spyderclass, redirect=None):
    self._redirect = redirect
    if redirect is None:
      assert spyderclass is None or \
       (hasattr(spyderclass, "typename") and hasattr(spyderclass.typename, "__call__"), spyderclass)
    self._spyderclass = spyderclass
    self._tree = general()
    self._exc_targets = []
    self._spyderobject = dummy()
    self._trig_targets = []

  def set_parent(self, parent):
    self.parent = parent
    
  def build(self, beename):
    self.beename = beename
  
  def set_spyderobject(self, spyderobject):
    red = self._get_redirect()
    if red: return red.set_spyderobject(spyderobject)
    assert isinstance(spyderobject, self._spyderclass)
    self._spyderobject = spyderobject
  
  def set_spydersubobject(self, pattern, subobject):
    if pattern == "_self": return self.set_spyderobject(subobject)
    
    red = self._get_redirect()
    if red: return red.set_spydersubobject(pattern, subobject)
    
    obj = self._spyderobject
    pat = pattern.split(".")
    for v in pat[:-1]:
      obj = getattr(obj, v)
    setattr(obj, pat[-1], subobject)  
  
  def get_spyderobject(self):
    red = self._get_redirect()
    if red: return red.get_spyderobject()
    obj = self._spyderobject
    if isinstance(obj, self._spyderclass): return obj
    obj = self._spyderclass(obj.__dict__)
    self._spyderobject = obj
    return obj
      
  def place(self): 
    self._get_redirect()
    def setter(spydergetter, attr, val):
      setattr(spydergetter(), attr, val)

    def getter0(spydergetter, attr):
      ret = getattr(spydergetter(), attr)
      return ret

    def getter(spydergetter, cls, attr):
      ret = getattr(spydergetter(), attr)
      if isinstance(ret, cls): return ret
      return cls(ret)

    def declare_subtree(self, subtree, spydertypetree, spydergetter):
      for mname, mtypetree in spydertypetree.members:
        mtypename = mtypetree.typename
        if hasattr(mtypetree,"arraycount"):
          for c in range(mtypetree.arraycount): mtypename += "Array"        
        mtype = getattr(Spyder,mtypename)

        catch = functools.partial(self.catch, mname)      
        i = ("bee","antenna",mname, mtypename)
        p = functools.partial(setter, spydergetter, mname) 
        pp = tryfunc(catch, p)
        libcontext.plugin(i, plugin_supplier(pp))
        o = ("bee","output",mname, mtypename)
        p = functools.partial(getter, spydergetter, mtype,mname) 
        pp = tryfunc(catch, p)
        libcontext.plugin(o, plugin_supplier(pp))

        if hasattr(mtypetree,"arraycount") and mtypetree.arraycount: continue
        nested = (hasattr(mtypetree,"members") and mtypetree.members is not None)
        if nested:
          c = libcontext.context(mname)
          libcontext.push(mname)        
          setattr(subtree,mname, general())
          msubtree = getattr(subtree, mname)
          mspydergetter = functools.partial(getter0,spydergetter, mname)
          declare_subtree(self, msubtree, mtypetree, mspydergetter)
          libcontext.pop()

    spydertypetree = None
    try:
      spydertypetree = self._spyderclass._typetree()
      assert isinstance(spydertypetree, spyder.core.typetreeclass)
    except AttributeError:
      pass


    libcontext.socket(("evexc","exception"), socket_container(self.add_exc_target))
    c = libcontext.context(self.beename)

    if self._spyderclass is not None:
      catch = functools.partial(self.catch, None)
      i = ("bee","antenna",self.beename, self._spyderclass.typename())
      p = functools.partial(setattr, self, "_spyderobject") 
      pp = tryfunc(catch, p)
      pp2a = plugin_supplier(pp)
      libcontext.plugin(i, pp2a)
      o = ("bee","output",self.beename, self._spyderclass.typename())
      p = self.get_spyderobject
      pp = tryfunc(catch, p)
      pp2b = plugin_supplier(pp)
      libcontext.plugin(o, pp2b)

    libcontext.push(self.beename)        

    if self._spyderclass is not None:
      i2 = ("bee","antenna","_self", self._spyderclass.typename())
      o2 = ("bee","output","_self", self._spyderclass.typename())
      libcontext.plugin(i2, pp2a)
      libcontext.plugin(o2, pp2b)
    
    i = ("bee","antenna","__trig__","trigger")
    p = self._trig 
    pp = tryfunc(catch, p)
    libcontext.plugin(i, plugin_supplier(pp))
    o = ("bee","output","__on_trig__", "trigger")
    s = self._add_trig_target
    libcontext.socket(o, socket_container(s))
    
    if spydertypetree is not None:
      declare_subtree(self, self._tree, spydertypetree, lambda: self._spyderobject)
    libcontext.pop()


class extern(blockbase):
  def __init__(self, spyderclass,redirect=None):
    self._redirect = redirect
    if redirect is None:
      assert spyderclass is None or \
       (hasattr(spyderclass, "typename") and hasattr(spyderclass.typename, "__call__"), spyderclass)
    self._spyderclass = spyderclass
    self._tree = general()
    self._exc_targets = []
    self._trig_targets = []
    self._getters = {}
    self._getter_types = {}
    self._subgetters_req = {}
    self._subgetters_opt = {}
    
  def set_parent(self, parent):
    self.parent = parent
    
    
  def build(self, beename):
    self.beename = beename 

  def _register_getter(self, pattern, func):  
    red = self._get_redirect()
    if red: return red._register_getter(pattern, func)    

    if pattern is None:
      if len(self._getters):
        p2 = list(self._getters.keys())[0]
        raise TypeError("Cannot define connection for %s: connection for %s has already been defined" % (str(pattern), str(p2)))          
    else:
      if None in self._getters:
        raise TypeError("Cannot define connection for %s: connection for the full object has already been defined" % str(pattern))
      for n in range(1,len(pattern)):
        p2 = pattern[:n]
        if p2 in self._getters:
          raise TypeError("Cannot define connection for %s: connection for %s has already been defined" % (str(pattern), str(p2)))
      for p2 in self._getters:
        if p2[:len(pattern)] == pattern:
          raise TypeError("Cannot define connection for %s: connection for %s has already been defined" % (str(pattern), str(p2)))                  

    self._getters[pattern] = func

  def _getter(self, pattern):
    red = self._get_redirect()
    if red: return red._getter(pattern)    

    #direct input
    if pattern is None:
      if None in self._getters: return self._getters[None]()
      err = AttributeError("Input not found")
    else:
      for n in reversed(range(-1,len(pattern))):
        if n == -1: subpattern = None
        else: subpattern = pattern[:n+1]
        if subpattern not in self._getters: continue
        ret = self._getters[subpattern]()
        rest = pattern[n+1:]
        for attr in rest:
          ret = getattr(ret, attr)
        return ret
      err = AttributeError("Input for %s not found" % str(pattern))

    #no direct input; try composite input
    ret = dummy()
    ok = True
    some = False
    if pattern not in self._subgetters_req:
      ok = False
    else:
      for subpattern, attr in self._subgetters_req[pattern]:          
        try:
          v = self._getter(subpattern)
          some = True
        except AttributeError:
          ok = False
          if some: break
          continue
        setattr(ret, attr, v)

      for subpattern, attr in self._subgetters_opt[pattern]:
        try:
          v = self._getter(subpattern)
          some = True          
        except AttributeError:          
          continue
        setattr(ret, attr, v)

    if ok:
      try:
        ret = self._getter_types[pattern](ret.__dict__)
      except AttributeError:
        ok = False

    if not ok: 
      if not some: 
        raise err
      else: #some inputs where present, but the object still couldn't be constructed          
        raise KeyError(pattern)

    return ret
       
  def place(self):    
    self._get_redirect()
      
    def declare_subtree(self, subtree, spydertypetree, pattern):
      pat2 = pattern
      if pat2 == (): pat2 = None
      self._subgetters_req[pat2] = []
      self._subgetters_opt[pat2] = []
      for mname, mtypetree in spydertypetree.members:
        subpattern = pattern + (mname,)
        if mtypetree.is_default: 
          self._subgetters_opt[pat2].append((subpattern, mname))
        else:
          self._subgetters_req[pat2].append((subpattern, mname))
        mtypename = mtypetree.typename
        if hasattr(mtypetree,"arraycount"):
          for c in range(mtypetree.arraycount): mtypename += "Array"
        mtype = getattr(Spyder,mtypename)
        self._getter_types[subpattern] = mtype

        catch = functools.partial(self.catch, mname)
        i = ("bee","antenna",mname, mtypename)
        s = functools.partial(self._register_getter, subpattern)
        ss = tryfunc(catch, s)
        libcontext.socket(i, socket_single_optional(ss))
        o = ("bee","output",mname, mtypename)
        p = functools.partial(self._getter, subpattern)
        pp = tryfunc(catch, p)
        libcontext.plugin(o, plugin_supplier(pp))

        if hasattr(mtypetree,"arraycount") and mtypetree.arraycount: continue
        nested = (hasattr(mtypetree,"members") and mtypetree.members is not None)
        if nested:
          c = libcontext.context(mname)
          libcontext.push(mname)        
          setattr(subtree,mname, general())
          msubtree = getattr(subtree, mname)
          declare_subtree(self, msubtree, mtypetree, subpattern)
          libcontext.pop()

    spydertypetree = None
    try:
      spydertypetree = self._spyderclass._typetree()
      assert isinstance(spydertypetree, spyder.core.typetreeclass)
    except AttributeError:
      pass


    libcontext.socket(("evexc","exception"), socket_container(self.add_exc_target))
    c = libcontext.context(self.beename)

    if self._spyderclass is not None:
      catch = functools.partial(self.catch, None)
      i = ("bee","antenna",self.beename, self._spyderclass.typename())
      s = functools.partial(self._register_getter, None)
      ss = tryfunc(catch, s)
      ss2 = socket_single_optional(ss)
      libcontext.socket(i, ss2)
      o = ("bee","output",self.beename, self._spyderclass.typename())
      p = functools.partial(self._getter, None)
      pp = tryfunc(catch, p)
      pp2 = plugin_supplier(pp)
      libcontext.plugin(o, pp2)

    libcontext.push(self.beename)
    if self._spyderclass is not None:
      i2 = ("bee","antenna","_self", self._spyderclass.typename())
      o2 = ("bee","output","_self", self._spyderclass.typename())
      libcontext.socket(i2, ss2)
      libcontext.plugin(o2, pp2)
    
    i = ("bee","antenna","__trig__","trigger")
    p = self._trig 
    pp = tryfunc(catch, p)
    libcontext.plugin(i, plugin_supplier(pp))
    o = ("bee","output","__on_trig__", "trigger")
    s = self._add_trig_target
    libcontext.socket(o, socket_container(s))

    self._getter_types[None] = self._spyderclass
    if spydertypetree is not None:
      declare_subtree(self, self._tree, spydertypetree, ())
    libcontext.pop()
