import libcontext, functools, traceback, sys
from .hivemodule import beehelper
from .resolve import resolve

class ConfigureBeeException(Exception): pass

class delayedcall(object):
  def __init__(self,parent,appendfunc, attrs, stack):
    self.parent = parent
    self.appendfunc = appendfunc
    self.attrs = attrs
    self.stack = stack
  def __call__(self, *args, **kargs):
    self.appendfunc(self.attrs, self.stack, *args, **kargs)
    return self.parent
  def __getattr__(self, attr):
    newattrs = list(self.attrs)
    newattrs.append(("getattr", attr))
    return delayedcall(self.parent, self.appendfunc,newattrs, self.stack)  
  def __getitem__(self, attr):
    newattrs = list(self.attrs)
    newattrs.append(("getitem", attr))
    return delayedcall(self.parent, self.appendfunc,newattrs, self.stack)  

class configure_base(beehelper):
  def getinstance(self): return self
  def configure(self, beedict):
    pass
  def bind(self): 
    pass    
  def place(self): 
    pass    
  def hive_init(self, beedict):
    pass
  def set_parameters(self,params):
    pass    
  
class configure(configure_base):
  def __init__(self, target):
    self.target_original = target
    self.configuration = []
  
    self.target = target
    self.bound_target = True
          
    if isinstance(target, str):
      self.target = target
    elif hasattr(target, "__get_beename__"):
      self.target = target
      self.bound_target = False
    else: raise TypeError(target)
    
  def bind(self):
    if not self.bound_target:
      if isinstance(self.target.__get_beename__, tuple): raise TypeError(self.target)
      self.target = self.target.__get_beename__(self.target)
      self.bound_target = True      

  def configure(self, beedict):
    if not self.bound_target:
      from .drone import drone
      if (isinstance(self.target.instance, drone) \
       or isinstance(self.target.instance, blockbase)) \
       and self.target.instance in beedict.values():
        self.target = [v for v in beedict if beedict[v] is self.target.instance][0]  
      else:        
        self.bind()
    n = beedict[self.target]
    if n is self: raise Exception("bee.configure target '%s' is self" % self.target)
    from .worker import workerframe
    if isinstance(n, workerframe):
      n = n.bee    
    for attrs, stack, args, kargs in self.configuration:
      args = tuple([resolve(a,parameters=self.parameters) for a in args])
      kargs = dict((a,resolve(kargs[a],parameters=self.parameters)) for a in kargs)
      try:
        nn = n       
        for mode, attr in attrs:
          if mode == "getattr":
            nn = getattr(nn, attr)
          elif mode == "getitem":
            nn = nn[attr]
          else: raise Exception(mode) #should never happen            
        nn(*args, **kargs)
      except Exception as e:
        s1 = traceback.format_list(stack[:-1])
        tbstack = traceback.extract_tb(sys.exc_info()[2])
        s2 = traceback.format_list(tbstack[1:])
        s3 = traceback.format_exception_only(type(e),e)
        s = "\n"+"".join(s1 + s2 + s3)
        raise ConfigureBeeException(s)
    if isinstance(n, configure_base): n.configure()
    
  def set_parameters(self, parameters):
    self.parameters = parameters
      
  def __configure_append__(self, attr, stack, *args, **kargs):
    self.configuration.append((attr, stack, args, kargs))
    return self

  def __getattr__(self, attr):
    if attr == "typename": raise AttributeError
    stack = traceback.extract_stack()  
    return delayedcall(self, self.__configure_append__, [("getattr",attr)], stack)
  def __getitem__(self, attr):
    stack = traceback.extract_stack()  
    return delayedcall(self, self.__configure_append__, [("getitem",attr)], stack)
    
class multiconfigure(configure_base):
  def __init__(self, *targets):
    self.targets = []
    for t in targets:           
      try:
        assert isinstance(t, configure_base)     
      except AssertionError:
        if not isinstance(t, list): raise
        t = [tt for tt in t]
        t = multiconfigure(*t)
      self.targets.append(t)
  def bind(self):
    for t in self.targets: t.bind()
  def configure(self, beedict):
    for t in self.targets: t.configure(beedict)
  def set_parameters(self,params):
    for t in self.targets: t.set_parameters(params)
  def hive_init(self, beedict):
    for t in self.targets: t.hive_init(beedict)
  
