import spyder, Spyder
spyder.load("basic")

from ..block import intern as block_intern, extern as block_extern
from .. import configure as bee_configure
from .blockbee_variable import get_variable, get_link, get_transistor, get_triggered

def get_pattern_type(pattern, typetree):
  dot = pattern.find(".")
  if dot == -1:
    try:
      ret = [t[1] for t in typetree.members if t[0] == pattern][0]
    except IndexError:
      print(typetree.members, pattern)
      raise
    ret = ret.typename
    ret = getattr(Spyder,ret)
    return ret
  else:
    head = pattern[:dot]
    subpattern = pattern[dot+1:]
    try:    
      subtree = [t[1] for t in typetree.members if t[0] == head][0]
    except IndexError:
      print(typetree.members, head)
      raise
    return get_pattern_type(subpattern, subtree)
  
def check_patterns(patterns):
  currpatterns = set()
  currpatterns2 = set()
  for p in patterns:
    if p == "_self" and len(currpatterns) > 0:
      raise TypeError("Cannot define connection for _self: connection for _self has already been defined" %  currpatterns[0])
    elif "_self" in currpatterns:
      raise TypeError("Cannot define connection for %s: connection for _self has already been defined" %  p)
    
    pp = tuple(p.split("."))
    for n in range(1, len(pp)):
      pp2 = pp[:n]
      if pp2 in currpatterns2:
        p2 = ".".join(pp2)
        raise TypeError("Cannot define connection for %s: connection for %s has already been defined" %  (p,p2))
    for pp2 in currpatterns2:
      if pp2[:len(pp)] == pp:
        p2 = ".".join(pp2)
        raise TypeError("Cannot define connection for %s: connection for %s has already been defined" %  (p,p2))

    currpatterns.add(p)
    currpatterns2.add(pp)

from .. import output as bee_output, antenna as bee_antenna, connect as bee_connect

def build_blockcon_input_extern(stub, blockname, con, typetree, sptype):  
  blockname = str(blockname)
  #for now, disalllow multi-level access: 
  # if x.y.z exists, x.y and x and _self cannot exist
  # eventually, this requirement will be relaxed for connectivities that were declared intern
  patterns = set([c.pattern for c in con]) #config blocks aren't allowed
  check_patterns(patterns)

  ret = []
  prim = 0
  for c in con:
    beename = stub + "_" + "_".join(c.pattern.split("."))
    if c.pattern == "_self": typ = sptype
    else: typ = get_pattern_type(c.pattern, typetree)
    value = None
    if c.pdata is not None:
      assert isinstance(c.pdata[0], typ), (c.pdata[0], typ)
      value = c.pdata[0]      
    if c.mode == "extern":
      beelink = get_link("pull",typ.typename())()
      beename2 = beename + "_link"      
      bee =  bee_antenna(beelink.inp)
    else: #intern or primary
      beelink = get_variable(typ.typename(), value)()
      beename2 = beename + "_var"      
      bee =  bee_antenna(beelink.inp)
          
    ret.append((beename2,beelink))
    ret.append((beename,bee))

    if c.pattern.find(".") == -1:
      target =  (blockname, c.pattern)        
    else:
      target =  (blockname,) + tuple(c.pattern.split("."))    
    con = bee_connect(beelink.outp,target)
    beename_con = beename + "_con"
    ret.append((beename_con,con))
    if c.mode == "primary":
      prim += 1
      beename_con2 = beename + "_con2"
      con2 = bee_connect(beelink.trig,(blockname, "__trig__"))
      ret.append((beename_con2,con2))
  if prim == 0:
    beename = stub + "__trig__"
    beelink = get_link("push","trigger")()
    beename2 = beename + "_link"      
    bee =  bee_antenna(beelink.inp)
    ret.append((beename2,beelink))
    ret.append((beename,bee))
    con = bee_connect(beelink.outp,(blockname, "__trig__"))
    beename_con = beename + "_con"
    ret.append((beename_con,con))
        
  return ret

def build_blockcon_input_intern(stub, blockname, con, typetree, sptype):  
  blockname = str(blockname)

  ret = []
  prim = 0
  for c in con:
    beename = stub + "_" + "_".join(c.pattern.split("."))
    if c.pattern == "_self": typ = sptype
    else: typ = get_pattern_type(c.pattern, typetree)
    if c.pdata is not None:
      assert isinstance(c.pdata[0], typ), (c.pdata[0], typ)
      beename_conf = beename + "_conf"
      conf = bee_configure(blockname)
      conf.set_spydersubobject(c.pattern, c.pdata[0])
      ret.append((beename_conf,conf))
    
    if c.mode == "config": continue
    
    #else: mode is intern or primary
    beelink = get_link("push",typ.typename())()
    beename2 = beename + "_link"      
    bee =  bee_antenna(beelink.inp)          
    
    ret.append((beename2,beelink))
    ret.append((beename,bee))

    if c.pattern.find(".") == -1:
      target =  (blockname, c.pattern)        
    else:
      target =  (blockname,) + tuple(c.pattern.split("."))    
    con = bee_connect(beelink.outp,target)
    beename_con = beename + "_con"
    ret.append((beename_con,con))
    if c.mode == "primary":
      prim += 1
      beetrig = get_triggered(typ.typename())()
      beename3 = beename + "_triggered"      
      ret.append((beename3,beetrig))
      beename_con2 = beename + "_con2"
      con2 = bee_connect(beelink.outp,beetrig.inp)
      ret.append((beename_con2,con2))
      beename_con3 = beename + "_con3"
      con3 = bee_connect(beetrig.trig,(blockname, "__trig__"))
      ret.append((beename_con3,con3))
  if prim == 0:
    beename = stub + "__trig__"
    beelink = get_link("push","trigger")()
    beename2 = beename + "_link"      
    bee =  bee_antenna(beelink.inp)
    ret.append((beename2,beelink))
    ret.append((beename,bee))
    con = bee_connect(beelink.outp,(blockname, "__trig__"))
    beename_con = beename + "_con"
    ret.append((beename_con,con))
        
  return ret

def build_blockcon_output(stub, blockname, con, typetree, sptype):  
  """
  It doesn't matter if the output block is declared intern or extern
  If data is provided, either the internal values of the block (intern) or 
   the hive elements that back it up (extern) are modified
  """
  blockname = str(blockname)

  ret = []
  prim = 0  
  for c in con:
    beename = stub + "_" + "_".join(c.pattern.split("."))
    if c.pattern == "_self": typ = sptype
    else: typ = get_pattern_type(c.pattern, typetree)

    if c.pdata is not None:
      assert isinstance(c.pdata[0], typ), (c.pdata[0], typ)    
      beename_conf = beename + "_conf"
      conf = bee_configure(blockname)
      conf.set_spydersubobject(c.pattern, c.pdata[0])
      ret.append((beename_conf,conf))
    
    if c.mode == "config": continue    
    
    if c.mode == "intern":
      beelink = get_link("pull",typ.typename())()
    elif c.mode == "primary":
      prim += 1    
      beelink = get_transistor(typ.typename())()
      con2 = bee_connect((blockname, "__on_trig__"), beelink.trig)
      beename_con2 = beename + "_con2"            
      ret.append((beename_con2,con2))
    beename2 = beename + "_link"      
    bee =  bee_output(beelink.outp)
    
    ret.append((beename2,beelink))
    ret.append((beename,bee))
    if c.pattern.find(".") == -1:
      source =  (blockname, c.pattern)        
    else:
      source =  (blockname,) + tuple(c.pattern.split("."))
    con = bee_connect(source, beelink.inp)
    beename_con = beename + "_con"
    ret.append((beename_con,con))
  if prim == 0:
    beename = stub + "__on_trig__"
    beelink = get_link("push","trigger")()
    beename2 = beename + "_link"      
    con = bee_connect((blockname, "__on_trig__"),beelink.inp)
    beename_con = beename + "_con"
    ret.append((beename_con,con))
    bee =  bee_output(beelink.outp)
    ret.append((beename2,beelink))
    ret.append((beename,bee))
    
  return ret

def build_blockcon_io(stub, stub2,  blockname, con, typetree, sptype):  
  blockname = str(blockname)

  ret = []
  for c in con:
    beename = stub + "_" + "_".join(c.pattern.split("."))
    bee2name = stub2 + "_" + "_".join(c.pattern.split("."))
    if c.pattern == "_self": typ = sptype
    else: typ = get_pattern_type(c.pattern, typetree)
    
    if c.pdata is not None:
      assert isinstance(c.pdata[0], typ), (c.pdata[0], typ)
      beename_conf = beename + "_conf"
      conf = bee_configure(blockname)
      conf.set_spydersubobject(c.pattern, c.pdata[0])
      ret.append((beename_conf,conf))
    
    if c.mode == "config": continue
    
    #else: mode is intern
    beelink = get_link("push",typ.typename())()
    beename2 = beename + "_link"      
    bee =  bee_antenna(beelink.inp)             
    ret.append((beename2,beelink))
    ret.append((beename,bee))

    bee2link = get_link("pull",typ.typename())()
    bee2name2 = bee2name + "_link2"      
    bee2 =  bee_output(bee2link.outp)             
    ret.append((bee2name2,bee2link))
    ret.append((bee2name,bee2))

    if c.pattern.find(".") == -1:
      target =  (blockname, c.pattern)        
    else:
      target =  (blockname,) + tuple(c.pattern.split("."))    

    con = bee_connect(beelink.outp,target)
    beename_con = beename + "_con"
    ret.append((beename_con,con))
    con2 = bee_connect(target, bee2link.inp)
    bee2name_con = bee2name + "_con"
    ret.append((bee2name_con,con2))
        
  return ret
           
def blockbee_build(bb):
  """
  Generates bee or combobee from blockbee
  if hivename is None: bee
  if hivename is None: combobee
  """
  #generation
  if bb.spyderclass is not None:
    class blockbee_spyderframe(spyderframe):
      for conv in bb.converters:
        bee.spyderhive.spyderhive.SpyderConverter(
          conv.inp,
          conv.outp,
          spyder.basic.import_link(conv.converter, False, globals())
        )
      del conv
    ret = blockbee_spyderframe
  elif bb.hivemap is not None:    
    from .hivemaphive import hivemapframe
    class blockbee_hivemapframe(hivemapframe):
      hivemap = bb.hivemap
      b,bl = None,None
      for b in bb.hivemap.blocks:      
        bl = block_intern
        if b.mode == "extern": bl = block_extern
        locals()[b.blockid] = bl(None, redirect = ("bees", "$hivemap", "bees", "$"+b.blockid))
      del b, bl
    ret = blockbee_hivemapframe
  elif bb.beefunc:
    beefunc = spyder.basic.import_link(bb.beefunc, False, globals())
    ret = beefunc()
  elif bb.beeobject:
    beeobj = spyder.basic.import_link(bb.beeobject, False, globals())
    ret = beeobj
  
  if len(bb.blocks):
    blockbee = ret
    class blockbeehive(blockbee):
      _blockname = None
      for _block in bb.blocks:
        _stub, _stub2 = None, None
        _spydertype = getattr(Spyder,_block.spydertype)
        _typetree = None
        if hasattr(_spydertype, "_typetree"):
          _typetree = _spydertype._typetree()
        _bees = []
        if _block.io == "input":
          _stub = "_inp_"+ _block.blockname
          if _block.mode == "extern":
            _bees = build_blockcon_input_extern(_stub, _block.blockname, _block.con, _typetree, _spydertype)
          elif _block.mode == "intern":
            _bees = build_blockcon_input_intern(_stub, _block.blockname, _block.con, _typetree, _spydertype)
        elif _block.io == "output":
          _stub = "_outp_"+ _block.blockname
          _bees = build_blockcon_output(_stub, _block.blockname, _block.con, _typetree, _spydertype)
        elif _block.io == "io":        
          _stub = "_inp_"+ _block.blockname        
          _stub2 = "_outp_"+ _block.blockname
          _bees = build_blockcon_io(_stub, _stub2, _block.blockname, _block.con, _typetree, _spydertype)
        elif _block.io == "config":
          _blockname = "_block_" + _block.blockname + "_config"
          locals()[_blockname] = bee_configure(_block.blockname)
          locals()[_blockname].set_spyderobject(_block.configdata[0])
          _beename, _bee = None, None
        for _beename, _bee in _bees:
          locals()[_beename] = _bee
        del _beename, _bee, _bees
      del _block, _blockname, _stub, _stub2, _spydertype, _typetree
    ret = blockbeehive
  return ret

def blockbee_make_bee_or_combo(bb):
  ret = blockbee_build(bb)()
  if bb.hivename is None: 
    return ret.getinstance()
  else:
    return [],{bb.hivename:[(bb.blockid,ret)]}
  
from ..combohive import combohive  
from .spyderhive import SpyderMethod, spyderhive, spyderframe

class blockbeehive(spyderhive):
  SpyderMethod("make_bee","Blockbee",blockbee_make_bee_or_combo)

class blockbeeframe(spyderframe):
  SpyderMethod("make_bee","Blockbee",blockbee_make_bee_or_combo)

class blockbeecombohive(combohive):
  SpyderMethod("make_combo","Blockbee",blockbee_make_bee_or_combo)
