#############################
### Skeleton animation drone
#############################

#In real life: 
#
#from some_library import compute_skeleton_animation_graph 
#
#now, just make a dummy function that multiplies its 3 inputs

def compute_skeleton_animation_graph(par1, par2, par3):
  print("computing skeleton from parameters:", par1, par2, par3)
  return par1*par2*par3

from bee import drone
import libcontext
from libcontext.socketclasses import *
from libcontext.pluginclasses import *

class skeleton_animator(drone):
  def __init__(self):
    self._cache = {}
  def get_cache(self): return self._cache
  def place(self):
    p = plugin_supplier(compute_skeleton_animation_graph)
    libcontext.plugin(("skeleton_animation", "compute"), p)
    p2 = plugin_supplier(self.get_cache)
    libcontext.plugin(("skeleton_animation", "get_cache"), p2)
   
#############################
### Skeleton animation worker
### (naive implementation)
#############################

from bee import drone, worker
from bee.segments import *

import libcontext
from libcontext.socketclasses import *
from libcontext.pluginclasses import *

class B1(worker):
  inp1 = antenna("pull", "float")
  inp2 = antenna("pull", "float")
  inp3 = antenna("pull", "float")

  b_inp1 = buffer("pull", "float")
  b_inp2 = buffer("pull", "float")
  b_inp3 = buffer("pull", "float")
  
  connect(inp1, b_inp1)
  connect(inp2, b_inp2)
  connect(inp3, b_inp3)
  
  outp = output("pull", "float")
  v_outp = variable("float")
  connect(v_outp, outp)
  
  @modifier
  def do_compute(self):
    par1 = self.b_inp1
    par2 = self.b_inp2
    par3 = self.b_inp3
    result = self.compute(par1, par2, par3)
    self.v_outp = result
  
  #When v_outp is requested for output, first query the antennas...
  pretrigger(v_outp, b_inp1, "output")
  pretrigger(v_outp, b_inp2, "output")
  pretrigger(v_outp, b_inp3, "output")
  #... then, trigger the "do_compute" modifier
  pretrigger(v_outp, do_compute, "output")
  
  def set_compute(self, compute):
    self.compute = compute
  def place(self):
    s = socket_single_required(self.set_compute)
    libcontext.socket(("skeleton_animation", "compute"), s)

#############################
### Skeleton animation worker
### (updates only if input changes)
#############################

from bee import drone, worker
from bee.segments import *

import libcontext
from libcontext.socketclasses import *
from libcontext.pluginclasses import *

class B2(worker):
  inp1 = antenna("pull", "float")
  inp2 = antenna("pull", "float")
  inp3 = antenna("pull", "float")

  b_inp1 = buffer("pull", "float")
  b_inp2 = buffer("pull", "float")
  b_inp3 = buffer("pull", "float")
  
  connect(inp1, b_inp1)
  connect(inp2, b_inp2)
  connect(inp3, b_inp3)
  
  outp = output("pull", "float")
  v_outp = variable("float")
  connect(v_outp, outp)
  
  old_input = variable(("float","float","float"))
  startvalue(old_input, (None,None,None))
  
  @modifier
  def do_compute(self):
    par1 = self.b_inp1
    par2 = self.b_inp2
    par3 = self.b_inp3
    if self.old_input == (par1, par2, par3):
      print("The input is the same as the old input, not recomputing")
      return
    result = self.compute(par1, par2, par3)
    self.old_input = par1, par2, par3
    self.v_outp = result
  
  #When v_outp is requested for output, first query the antennas...
  pretrigger(v_outp, b_inp1, "output")
  pretrigger(v_outp, b_inp2, "output")
  pretrigger(v_outp, b_inp3, "output")
  #... then, trigger the "do_compute" modifier
  pretrigger(v_outp, do_compute, "output")
  
  def set_compute(self, compute):
    self.compute = compute
  def place(self):
    s = socket_single_required(self.set_compute)
    libcontext.socket(("skeleton_animation", "compute"), s)

#############################
### Skeleton animation worker
### uses the cache stored on the drone
### only recomputes if the value is not in cache
#############################

from bee import drone, worker
from bee.segments import *

import libcontext
from libcontext.socketclasses import *
from libcontext.pluginclasses import *

class B3(worker):
  inp1 = antenna("pull", "float")
  inp2 = antenna("pull", "float")
  inp3 = antenna("pull", "float")

  b_inp1 = buffer("pull", "float")
  b_inp2 = buffer("pull", "float")
  b_inp3 = buffer("pull", "float")
  
  connect(inp1, b_inp1)
  connect(inp2, b_inp2)
  connect(inp3, b_inp3)
  
  outp = output("pull", "float")
  v_outp = variable("float")
  connect(v_outp, outp)
    
  @modifier
  def do_compute(self):
    par1 = self.b_inp1
    par2 = self.b_inp2
    par3 = self.b_inp3

    cache = self.get_cache()
    if (par1, par2, par3) in cache:
      print("The input was found in cache, not recomputing")
      return
    result = self.compute(par1, par2, par3)
    cache[(par1, par2, par3)] = result
    self.v_outp = result
  
  #When v_outp is requested for output, first query the antennas...
  pretrigger(v_outp, b_inp1, "output")
  pretrigger(v_outp, b_inp2, "output")
  pretrigger(v_outp, b_inp3, "output")
  #... then, trigger the "do_compute" modifier
  pretrigger(v_outp, do_compute, "output")
  
  def set_compute(self, compute):
    self.compute = compute
  def set_get_cache(self, get_cache):
    self.get_cache = get_cache
  def place(self):
    s = socket_single_required(self.set_compute)
    libcontext.socket(("skeleton_animation", "compute"), s)
    s2 = socket_single_required(self.set_get_cache)
    libcontext.socket(("skeleton_animation", "get_cache"), s2)

#############################
### Main hive
#############################

from bee import inithive, connect
import dragonfly.std

class mainhive(inithive):
  skeleton_animator()
  a = dragonfly.std.buffer("pull", "float")(None)
  
  #pick any:
  #b = B1()
  #b = B2()
  b = B3()
   
  c = dragonfly.std.variable("float")(None)
  d = dragonfly.std.variable("float")(None)
  e = dragonfly.std.variable("float")(None)
  
  connect(b, a)
  connect(c, b.inp1)
  connect(d, b.inp2)
  connect(e, b.inp3)

  trigger_a = dragonfly.std.trigger()
  connect(trigger_a, a)
  
main = mainhive()
m = main.getinstance()
m.build("m")
m.place()
m.close()
m.init()

################
# Do some computation
################

m.c.value = 1.0
m.d.value = 2.0
m.e.value = 3.0

#trigger worker A to request its input
m.trigger_a.trigger() 
print m.a.value

#trigger again, C-D-E unchanged
m.trigger_a.trigger() 
print m.a.value

m.c.value = 9.0
#trigger again, C has now changed
m.trigger_a.trigger() 
print m.a.value


