from __future__ import print_function

import dragonfly
import dragonfly.consolehive
import dragonfly.std, dragonfly.io, dragonfly.sys

import bee
from bee.pin import pinworker, inputpin, outputpin
from bee.pin import mediator, pinconnect, disconnect, remove, create

import libcontext
libcontext.use_networks = False

#Definition of six pin workers
class int_variable(pinworker):
  i = inputpin("pull","int")
  o = outputpin("pull","int")
  def run(self):
    self.o.value = self.i.value

class connector(pinworker):
  i = inputpin("push","trigger")
  o = outputpin("push","trigger")
  def run(self):
    self.o.value = self.i.value

class counter(pinworker):
  i1 = inputpin("push", "trigger")
  i2 = inputpin("pull", "int")
  o = outputpin("pull", "bool",refresh=False)(False)  
  counter = outputpin("pull", "int",refresh=False)(0)  
  def run(self):  
    if self.i1.value: 
      self.counter.value += 1
      if self.counter.value >= self.i2.value: 
        self.o.value = True
      else: 
        self.o.value = False
      print("COUNTER", self.beename, "%d/%d" % (self.counter.value, self.i2.value), self.o.value)


class simplecounter(pinworker):
  i = inputpin("push", "trigger")
  counter = outputpin("pull", "int",refresh=False)(0)  
  def run(self):  
    if self.i.value: 
      self.counter.value += 1

class reporter(pinworker):
  i = inputpin("pull", "int")
  def run(self):
    print("REPORT:", self.i.value,end="")

class andworker(pinworker):
  i1 = inputpin("pull", "bool")
  i2 = inputpin("pull", "bool")
  o = outputpin("push", "trigger")
  def run(self): 
    if self.i1.value and self.i2.value: 
      self.o.value = True
    else:
      self.o.value = False

from bee import connect
class main(dragonfly.consolehive.consolehive):
  ### exception handling
  raiser = bee.raiser()
  connect("evexc", raiser)

  ### pin drones  
  mediator()
  pinconnect = pinconnect()
  disconnect = disconnect()
  remove = remove()
  create = create()

  ### game logic
  t = dragonfly.io.ticksensor()
  tt = connector()
  connect(t,tt)

  x = simplecounter()  #x comes after t
  #connect(tt, x)      #added dynamically, after 10 ticks
  
  xx = reporter()      #xx comes after x; removed dynamically, after 20 ticks
  #connect(x, xx)      #added dynamically; auto-removed after 20 ticks
  
  v1 = dragonfly.std.variable("int")(10)
  
  var1 = int_variable()
  connect(v1, var1)
  var2 = int_variable()
  connect(v1, var2)
  
  k1 = dragonfly.io.keyboardsensor_trigger("SPACE")  
  c1 = counter()
  connect(k1, c1)
  #connect(var1, c1)  #added dynamically

  k2 = dragonfly.io.keyboardsensor_trigger("RETURN")  
  c2 = counter()
  connect(k2, c2)
  #connect(var2, c2)  #added dynamically

  #a = andworker()    #created dynamically
  #connect(c1, a.i1)  #added dynamically
  #connect(c2, a.i2)  #added dynamically

  con = connector()
  #connect(a, con)    #added dynamically

  e = dragonfly.sys.exitactuator()
  connect(con, e)

m = main().getinstance()
m.build("main")
m.place()
m.close()
m.init()

print("Demo hive")
m.pinconnect(m.var1.o, m.c1.i2)
print("OK")
m.pinconnect(m.var2.o, m.c2.i2)
m.pinconnect(m.x.counter, m.xx.i)

print("For the first 10 ticks, the simplecounter is not connected to the tick sensor")
for n in range(10):
  print("tick,",end=""); m.pacemaker.tick(); print()
m.pinconnect(m.tt.o, m.x.i)

print("The simplecounter increases every tick")
for n in range(10):
  print("tick,",end=""); m.pacemaker.tick(); print()

print("The connection is broken again")
m.disconnect(m.tt.o, m.x.i)
for n in range(10):
  print("tick,",end=""); m.pacemaker.tick(); print()

print("The reporter is removed altogether")
m.remove(m.xx)
for n in range(10):
  print("tick,",end=""); m.pacemaker.tick(); print()

print("An andworker is created dynamically")
m.a = m.create(andworker(), "a")
m.pinconnect(m.c1.o, m.a.i1)
m.pinconnect(m.c2.o, m.a.i2)
m.pinconnect(m.a.o, m.con.i)

print("Press 10x SPACE and 10x RETURN to quit (or press Esc)")
m.run()
