from __future__ import print_function, absolute_import

import os
import spyder, Spyder

class HivemapManager(object):
  def __init__(self, mainwin, workermanager, workerinstancemanager, file_dialog):
    self._mainwin = mainwin
    self._wim = workerinstancemanager
    self._workermanager = workermanager
    self._file_dialog = file_dialog

    self._lastsave = None
    
    win = self._mainwin.h()
    win.add_menu("&File")
    win.add_menu_action(
     "&File",
     "&New",
     self.clear,
     shortcut = "Ctrl+N",
     statustip = "Close the current Hivemap file and open a new one",
    )
    win.add_menu_action(
     "&File",
     "&Open",
     self.menu_load,
     shortcut = "Ctrl+O",
     statustip = "Open a Hivemap file",
    )
    win.add_menu_action(
     "&File",
     "&Save",
     self.menu_save,
     shortcut = "Ctrl+S",
     statustip = "Save the current Hivemap file",
    )
    win.add_menu_action(
     "&File",
     "Save as",
     self.menu_save_as,
     statustip = "Save the current Hivemap file in the desired location",
    )
    win.add_menu_action(
     "&File",
     "&Quit",
     win.close,
     shortcut = "Ctrl+Q",
     statustip = "Quits the editor",
    )

  """
  def _rebuild_evio_workers(self, hivemap):
    evio = ("evin", "evout", "everr", "evexc")
    ids = set()
    for con in hivemap.connections:
      if con.start.io in evio: 
        ids.add(con.start.workerid)
      if con.end.io in evio: 
        ids.add(con.end.workerid)
    types = set()
    for w in hivemap.workers:
      if w.workerid in ids:
        types.add(w.workertype)
    for workertype in types:
      self._workermanager.remove_built_worker(workertype)
      self._workermanager.build_worker(workertype, append_evio = True)
  """    
  def _load(self, hivemap):
    for w in hivemap.workers:
      x, y = w.position.x, w.position.y
      metaparamvalues = None
      if w.metaparameters is not None:
        metaparamvalues = {}
        for param in w.metaparameters:
          metaparamvalues[param.pname] = param.pvalue
      paramvalues = None
      if w.parameters is not None:
        paramvalues = {}
        for param in w.parameters:
          paramvalues[param.pname] = param.pvalue
      try:
        self._workermanager.instantiate(
         w.workerid, w.workertype, x, y, metaparamvalues, paramvalues
        )
      except KeyError:
        print("Unknown worker:", w.workertype, w.workerid.split(".")[-1])
        continue  
    
    count = 0
    for con in hivemap.connections:
      count += 1
      con_id = "con-%d" % count
      interpoints = [(ip.x, ip.y) for ip in con.interpoints]
      self._wim.add_connection(
        con_id,
        (con.start.workerid, con.start.io),
        (con.end.workerid, con.end.io),
        interpoints,    
      )
  
  def load(self, hivemapfile):
    self.clear()
    
    workermanager = self._workermanager
    localdir = os.path.split(hivemapfile)[0]
    workermanager.find_local_workers(localdir)
    workermanager.build_workers(local=True)

    hivemap = Spyder.Hivemap.fromfile(hivemapfile)
    self._load(hivemap)

  def save(self, hivemapfile):
    workermanager = self._workermanager
    workers, connections = [], []
    for workerid in sorted(workermanager.workerids()):
      node, mapping = self._wim.get_node(workerid)
      if node.empty: continue
      workertype, params, metaparams = workermanager.get_parameters(workerid)
      if params is not None: params = params.items()
      if metaparams is not None: metaparams = metaparams.items()
      w = Spyder.Worker (
        workerid,
        workertype,
        node.position,
        parameters = params,
        metaparameters = metaparams,
      )
      workers.append(w)
    for c in self._wim.get_connections():  
      start_node, start_mapping = self._wim.get_node(c.start_node)
      end_node, end_mapping = self._wim.get_node(c.end_node)
      start_attribute = start_mapping._outmapr[c.start_attribute]
      end_attribute = end_mapping._inmapr[c.end_attribute]
      con = Spyder.WorkerConnection (
        (c.start_node, start_attribute),
        (c.end_node, end_attribute),
        c.interpoints
      )
      connections.append(con)
    hivemap = Spyder.Hivemap(workers, connections)
    hivemap.tofile(hivemapfile)
    self._lastsave = hivemapfile

  def menu_save_as(self):
    hivemapfile = self._file_dialog("save")
    self.save(hivemapfile)  

  def menu_save(self):
    if self._lastsave is None:
      return self.menu_save_as()
    self.save(self._lastsave)  

  def menu_load(self):
    hivemapfile = self._file_dialog("open")
    self.load(hivemapfile)  
    self._lastsave = hivemapfile

  def clear(self):
    self._workermanager.clear_workers()
    self._workermanager.unbuild_local_workers()
