from __future__ import print_function, absolute_import

import spyder.formtools
import spyder.qtform
from spyder.qtform.anyQt.QtCore import QBuffer
from spyder.qtform.anyQt.QtUiTools import QUiLoader
from spyder.qtform import qtview, reserved
from spyder.formtools import model, controller, comparator, mvc
import functools

from ..params import typemap

loader = None

from . import pythoncode

def _PButtonCallback(callback, model, contains_reserved):
  val = model.get()
  if val is not None and contains_reserved:
    val = val.copy()
    for k in reserved:
      kk = k + "_"
      if kk in val:
        val[k] = val[kk]
        del val[kk]
  callback(val)

def remap_reserved(callback, params):
  params2 = params.copy()
  for k in reserved:
    kk = k+"_"
    if kk in params2:
      params2[k] = params2[kk]
      del params2[kk]
  callback(params2)

  
def PGenerator(paramnames, paramtypelist, paramvalues, 
  update_callback, buttons = []
 ):
  global loader
  if loader is None: loader = QUiLoader()

  contains_reserved = False
  paramnames2 = paramnames

  for k in reserved:
    if k in paramnames:
      contains_reserved = True
      paramnames2 = [p if p != k else k + "_" for p in paramnames2]

  params = list(zip(paramnames2, paramtypelist))
  #print("PARAMS", params)
  typetree = spyder.formtools.generate_typetree(params,typemap)
  #print("TYPETREE", typetree)
  form = spyder.core.spyderform(typetree)  
  #print("SUBFORM", form.type_, form.type_.__dict__)
  for pname,pvalue in paramvalues.items():
    pname2 = pname
    if pname in reserved: pname2 = pname + "_"
    if pvalue is not None:
      form._members[pname2].value = pvalue
  for buttonname, buttoncallback in buttons:
    form.add_button(buttonname, "after")
       
  m = model(typetree)
  con = controller(form, m)
  com = comparator(form)

  xml = spyder.qtform.xml(form=form)
  buf = QBuffer()
  buf.open(QBuffer.ReadWrite)
  buf.write(xml)
  buf.close()
  parwidget = loader.load(buf)

  v = qtview.qtview(form)
  v._wrap(parwidget)

  for pname in paramnames:
    pname2 = pname
    if pname in reserved: pname2 = pname + "_"
    f = form._members[pname2]  
    if hasattr(f, "type") and f.type == "pythoncode":
      widget = getattr(v, pname2).widget
      i = id(widget)
      if i in pythoncode.callbacks: continue
      pythoncode.callbacks.add(i)
      cb = functools.partial(pythoncode.highlight, widget)
      widget.textChanged.connect(cb)
      pythoncode.set_css(widget)

  #construct mvc
  my_mvc = mvc(form, lambda:m, lambda: v, lambda:con, lambda:com)

  #bind mvc
  my_mvc.bind()
  my_mvc.sync_m2v()
  my_mvc.sync_v2m()

  for pname in paramnames:
    pname2 = pname
    if pname in reserved: pname2 = pname + "_"
    f = form._members[pname2]  
    if hasattr(f, "type") and f.type == "pythoncode":
      widget = getattr(v, pname2).widget
      code = None
      if hasattr(f, "value"): code = f.value
      elif hasattr(f, "default"): code = f.default
      if code is not None:
        widget.setPlainText(code)
        pythoncode.highlight(widget)

  for buttonindex,button in enumerate(buttons):
    buttonname, buttoncallback = button
    cb = functools.partial(_PButtonCallback, buttoncallback, m, contains_reserved)
    v.buttons[buttonindex].listen(cb)

  up = update_callback
  if contains_reserved: up = functools.partial(remap_reserved, update_callback)
  m.listen(up)
  
  return parwidget, my_mvc
