from __future__ import print_function, absolute_import
import sys, os

import spyder
from . import HGui, PGui, HQt, PQt
HGui.init(HQt)
PGui.init(PQt)
from . import HivemapManager, Clipboard
from .PGui import PWindow, PWidgetWindow, PManager, PWorkerCreator, PControllerGeneral
from .worker import WorkerManager, WorkerBuilder, WorkerInstanceManager

def run(readmetxt = None):
  app = HGui.Application(sys.argv)
  from .HQt.anyQt import QtGui

  mainWin = HGui.MainWindow()
  clipboard = Clipboard()
  statusbar = HGui.StatusBar(mainWin)
  canvas = HGui.NodeCanvas(mainWin, clipboard, statusbar)
  controller_general = PGui.PControllerGeneral()
  pwins = {}
  pwins["general"] = PWindow(mainWin, "props-general", controller_general)
  pwins["params"] = PWidgetWindow(mainWin, "props-params")
  pwins["metaparams"] = PWidgetWindow(mainWin, "props-metaparams")
  pmanager = PManager(pwins)
  pwc = PWorkerCreator(mainWin, clipboard)
  mainWin.show()

  currdir = os.path.abspath(os.path.dirname(sys.argv[0]))
  workerbuilder = WorkerBuilder()
  workerinstancemanager = WorkerInstanceManager(canvas)
  controller_general.set_workerinstancemanager(workerinstancemanager)
  workermanager = WorkerManager(
   workerbuilder, workerinstancemanager, pmanager, pwc
  )
  controller_general.set_workermanager(workermanager)
  clipboard.set_workermanager(workermanager)
  workermanager.find_global_workers(currdir)
  workermanager.build_workers(local=False)

  hivemapmanager  = HivemapManager(
   mainWin, workermanager, workerinstancemanager, HGui.FileDialog
  )

  if len(sys.argv) > 1:
    #hivemapfile = "../manual/movingpanda/pandajump.web"
    hivemapfile = sys.argv[1]
    hivemapmanager.load(hivemapfile)
    hivemapmanager._lastsave = hivemapfile

  if readmetxt is not None:
    dialog = QtGui.QDialog()
    dialog.setMinimumSize(600,800)
    dialog.setModal(False)
    pte = QtGui.QPlainTextEdit(readmetxt, dialog)
    pte.setMinimumSize(600,800)
    dialog.show()

  app.h().mainloop()

