from __future__ import print_function, absolute_import

import sys, os

def add_worker(worker, workername, workers, metaworkers, found_workers):
  w = worker
  try:
    if hasattr(w, "guiparams") and ("antennas" in w.guiparams or "outputs" in w.guiparams):            
      workers[workername] = w
      found_workers.add(id(w))
      return True
    elif hasattr(w, "metaguiparams"):        
      metaworkers[workername] = w
      found_workers.add(id(w))
      return True
  except TypeError:
    pass
  return False

def find_workers(mod, modname, done_mods, found_workers, star = False, search_path=True):
  if id(mod) in done_mods: return {},{}
  import glob,os,imp
  workers = {}
  metaworkers = {}
  done_mods.add(id(mod))
  modname2 = modname.rstrip(".__init__")
  suspect = []
  for v in dir(mod):
    if v in ("frame","hive","closedhive","inithive","raiser"): continue
    if isinstance(v, os.__class__): #is a module...
      newmodname = modname+"."+v
      newworkers, newmetaworkers, newsuspect = \
       find_workers(v, newmodname, done_mods, found_workers)
      workers.update(newworkers)
      metaworkers.update(newmetaworkers)
      suspect += newsuspect
      continue
    vv = getattr(mod, v)
    if id(vv) in found_workers: continue
    workername = modname+"."+v
    if hasattr(vv, "__module__"): 
      if not vv.__module__.startswith(modname2):
        if hasattr(vv, "guiparams") or hasattr(vv, "metaguiparams"):
          if vv.__module__ != "bee.drone" and not vv.__module__.startswith("bee.segments"):
            suspect.append((workername, v, vv))
        continue
    add_worker(vv, workername, workers, metaworkers, found_workers)
  path = os.path.split(mod.__file__)[0]
  newmodules = []
  newstarmodules = []
  if search_path:
    for f in os.listdir(path):
      ff = path + os.sep + f
      if star:
        if os.path.isfile(ff) and ff.endswith(".py"):
          n = f[:-len(".py")]
          newstarmodules.append((n,ff))
      else:
        if os.path.isdir(ff) and os.path.exists(ff+os.sep+"__init__.py"): newmodules.append(f)
  for n in newmodules:
    newmodname = modname + "." + n
    if newmodname in workers or newmodname in metaworkers: continue
    
    try:
      __import__(newmodname)
    except ImportError:    
      print("Error in importing", newmodname)
      import traceback
      traceback.print_exc()
      continue
    newmod = sys.modules[newmodname]
    newworkers, newmetaworkers, newsuspect = \
     find_workers(newmod, newmodname, done_mods, found_workers)
    workers.update(newworkers)
    metaworkers.update(newmetaworkers)
    suspect += newsuspect
  for n,ff in newstarmodules:
    newmodname = modname + "." + n
    if newmodname in workers or newmodname in metaworkers: continue
    try:
      fil = open(ff,"r")
      imp.load_module(newmodname,fil,ff,("py","r",imp.PY_SOURCE ))
    except ImportError:    
      print("Error in importing", newmodname)
      import traceback
      traceback.print_exc()
      continue
    newmod = sys.modules[newmodname]
    newworkers, newmetaworkers, newsuspect = find_workers(
     newmod, newmodname,done_mods,found_workers, search_path=False
    ) 
    workers.update(newworkers)
    metaworkers.update(newmetaworkers)
    suspect += newsuspect
    
  return workers, metaworkers, suspect


class WorkerFinder(object):
  @staticmethod
  def find_locationmodules(locationconfs, localdir=None):
    locationmodules = [] 
    
    syspath = []
    if localdir is not None: syspath = [localdir]
    syspath += list(sys.path)
    
    for lconf in locationconfs:
      for l in open(lconf).readlines():
        l = l.strip()
        if l.startswith("#"): continue
        if not len(l): continue
        if l.startswith("@"): 
          if localdir is not None:
            syspath.append(localdir+os.sep+l[1:])
          else: 
            syspath.append(l[1:])
        else: locationmodules.append(l)
    return locationmodules, syspath

  def __init__(self, locationmodules, syspath, found_workers = None):
    self._done_mods = set()
    self._found_workers = set()
    if found_workers is not None:
      for w in found_workers: 
        self._found_workers.add(w)

    workers = {}
    metaworkers = {}
    suspect = []
    syspath_backup = list(sys.path)
    sys.path = syspath 
    for m in locationmodules:
      star = False
      if m.endswith(".*"):
        star = True
        m = m[:-2]
      try:
        __import__(m)
        mod = sys.modules[m]
        newworkers, newmetaworkers, newsuspect = find_workers(
         mod, m, self._done_mods, self._found_workers, star
        )
          
        workers.update(newworkers)
        metaworkers.update(newmetaworkers)
        suspect += newsuspect
      except:
        print("Error in importing", m)
        import traceback
        traceback.print_exc()

    sys.path = syspath_backup
    
    for workername, varname, worker in suspect:
      if id(worker) in self._found_workers: continue
      altworkername = worker.__module__ + "." + varname
      if altworkername in workers or altworkername in metaworkers:
        continue
      add_worker(
       worker, workername, workers, metaworkers, self._found_workers
      )
          
    self.workers = workers
    self.metaworkers = metaworkers
  def get_metaworker(self, metaworkername):
    metaworker = self.metaworkers[metaworkername]
    return metaworker, metaworker.metaguiparams
