def workerid_to_nodename(workerid):
  return workerid.split(".")[-1]

class WorkerInstance(object):
  _morphing = {
    "evin":  {"default":"default_evio","plain":"plain_evio"},
    "evout": {"default":"default_evio","plain":"plain_evio"},
    "everr": {"default":"default_evio","plain":"plain_evio"},
    "evexc": {"default":"default_evio","plain":"plain_evio"},
  }
  def __init__(self, type_, profiles, paramnames, paramtypelist ):
    self.type = type_
    self.profiles = profiles
    self.paramnames = paramnames
    self.paramtypelist = paramtypelist
    self.curr_profile = None
  def _get_morph(self, attribute):
    if attribute not in self._morphing: return None
    m = self._morphing[attribute]
    if self.curr_profile not in m: return None
    return m[self.curr_profile]
  def profile(self):
    return self.profiles[self.curr_profile]
  def check_antenna(self, attribute):
    prof = self.profiles[self.curr_profile]
    attribs, mapping = prof
    at = mapping.inmap[attribute]
    for a in attribs:
      if a.name == at:
        if a.inhook is not None:
          return None #OK
        else:
          break  
    ret = self._get_morph(attribute)
    if ret is None:
      raise KeyError(attribute)
    return ret  
  def check_output(self, attribute):
    prof = self.profiles[self.curr_profile]
    attribs, mapping = prof
    at = mapping.outmap[attribute]
    for a in attribs:
      if a.name == at:
        if a.outhook is not None:
          return None #OK
        else:
          break  
    ret = self._get_morph(attribute)
    if ret is None:
      raise KeyError(attribute)
    return ret  

from ..HGui import Node

class WorkerInstanceManager(object):
  def __init__(self, canvas):
    self._canvas = canvas
    self.observers_selection = canvas.observers_selection
    self.observers_remove = canvas.observers_remove
    self._workerinstances = {}
    self._workerparams = {}
  def add_workerinstance(self, workerid, workerinstance,x,y):
    assert workerid not in self._workerinstances, workerid
    attribs, mapping = workerinstance.profiles["default"]
    nodename = workerid_to_nodename(workerid)
    node = Node(nodename, (x,y), attribs)
    self._canvas.add_node(workerid, node)
    workerinstance.curr_profile = "default"
    self._workerinstances[workerid] = workerinstance
  def select(self, workerids):
    self._canvas.select(workerids)
  def morph_worker(self, workerid, morph):
    wi = self._workerinstances[workerid]
    
    m0 = wi.profile()
    m1 = wi.profiles[morph]
    map0, map1 = m0[1], m1[1]
    maps = []
    for at in "in", "out":
      cmap = {}
      mapr = getattr(map0,"_"+at+"mapr")
      mapnew = getattr(map1,"_"+at+"map")
      for v,k in mapr.items():
        if k is None: continue
        if mapnew[k] is None: continue 
        vv = mapnew[k]
        cmap[v] = vv
      maps.append(cmap)
    nodename = workerid_to_nodename(workerid)
    node0 = self._canvas.get_node(workerid)
    x,y = node0.position
    newnode = Node(nodename, (x,y), m1[0])      
    self._canvas.morph_node(workerid, newnode, maps[0], maps[1])
    wi.curr_profile = morph
    if workerid in self._workerparams:
      self.set_parameters(workerid, self._workerparams[workerid])
  def add_connection(self, _id, start, end, interpoints = []):
    start_id, start_attr = start
    wi_start = self._workerinstances[start_id]
    morph = wi_start.check_output(start_attr)
    if morph is not None: self.morph_worker(start_id, morph)
    
    end_id, end_attr = end
    wi_end = self._workerinstances[end_id]
    morph = wi_end.check_antenna(end_attr)
    if morph is not None: self.morph_worker(end_id, morph)
    
    map_start = wi_start.profile()[1].outmap
    map_end = wi_end.profile()[1].inmap
    startm = (start[0], map_start[start[1]])
    endm = (end[0], map_end[end[1]])
    self._canvas.add_connection(_id, startm, endm, interpoints)
  def get_node(self, workerid):
    node = self._canvas.get_node(workerid)
    try:
      wi = self._workerinstances[workerid]
      mapping = wi.profile()[1]
    except KeyError: #empty
      mapping = None  
    return node, mapping
  def get_workerinstance(self, workerid):
    return self._workerinstances[workerid]

  def get_connections(self):
    return self._canvas.get_connections()
  def add_empty(self, workerid, x, y):
    nodename = workerid_to_nodename(workerid)
    node = Node(nodename, (x,y), [], empty = True)
    self._canvas.add_node(workerid, node)
  def gui_removes_workerinstance(self, workerid):
    workerinstance = self._workerinstances.pop(workerid)
    self._workerparams.pop(workerid, None)
  def remove_workerinstance(self, workerid):
    workerinstance = self._workerinstances.pop(workerid)
    self._canvas.remove_node(workerid)
    self._workerparams.pop(workerid, None)
  def remove_empty(self, workerid):
    self._canvas.remove_node(workerid)
  def rename_workerinstance(self, old_workerid, new_workerid):
    new_nodename = workerid_to_nodename(new_workerid)
    self._canvas.rename_node(old_workerid, new_workerid, new_nodename)  
    inst = self._workerinstances.pop(old_workerid)
    self._workerinstances[new_workerid] = inst
  def set_parameters(self, workerid, params):
    self._workerparams[workerid] = params
    wi = self._workerinstances[workerid]
    pmapping = wi.profile()[1].pmap
    mparams = {}
    for p in params:
      mp = pmapping[p]
      if mp is None: continue
      mparams[p] = params[mp]
    for p,v in mparams.items():
      self._canvas.set_attribute_value(workerid, p, v)
