networks = {}
_detect_networks = set()

from .context import bee_evexc, namesortfunc

class Network(object):
  def __init__(self, contexts):
    self.contexts = sorted(contexts,key=lambda c: namesortfunc(c.contextname))
    self.plugins = {}
    self.sockets = {}
    self.socket_imports = []
    self.plugin_imports = []
    self.socket_imps = set()
    self.plugin_imps = set()
    self.orig = None
    self.resolved = False
    self.plugaliases = {}
    self.sockaliases = {}
  def import_socket(self, contextinstance, name, newname = None, optional = False):  
    from .context import context as contextclass
    assert isinstance(contextinstance, contextclass)
    assert self.orig is not None
    self.resolved = False
    if newname == None: newname = name
    if contextinstance in self.contexts:
      if newname == name: return
      contextinstance = self.contexts[-1]
    n = (self.orig, contextinstance, name, newname, optional)
    if n[1:] not in self.socket_imps:    
      self.socket_imports.append(n)
      self.socket_imps.add(n[1:])
  def import_plugin(self, contextinstance, name, newname = None, optional = False):  
    from .context import context as contextclass
    assert isinstance(contextinstance, contextclass)
    assert self.orig is not None
    self.resolved = False
    if newname == None: newname = name
    if contextinstance in self.contexts:
      if newname == name: return
      contextinstance = self.contexts[-1]
    n = (self.orig, contextinstance, name, newname, optional)
    if n[1:] not in self.plugin_imps:    
      self.plugin_imports.append(n)        
      self.plugin_imps.add(n[1:])        
  def resolve(self):
    if self.resolved: return
    #print("RESOLVE", self.contexts[-1].contextname)
    from .context import retrieve_sockets, retrieve_plugins
    for sp in "socket","plugin":
      todo = []    
      if sp == "socket":
        sockplug = self.sockets
        imports = self.socket_imports
        retrieve = retrieve_sockets
        aliases = self.sockaliases
      else:
        sockplug = self.plugins
        imports = self.plugin_imports
        retrieve = retrieve_plugins
        aliases = self.plugaliases
      #print sp, sockplug.keys()
 
      for orig, contextinstance, name, newname, optional in imports:
        if contextinstance in self.contexts:
          if name in aliases: name = aliases[name]
          if newname in aliases: newname = aliases[newname]
          if newname == name: continue          
          td = ("alias",(name, newname, optional, contextinstance, orig))
          todo.append(td)
        else:
          td = ("import",(contextinstance, name, newname, optional, orig))
          todo.append(td)
      contains = set(sockplug.keys())
      for action, param in todo:
        if action == "alias":
          name, newname, optional, orig_source, orig_target = param
          if name in contains: contains.add(newname)
      for action, param in todo:
        if action == "alias":
          name, newname, optional, orig_source, orig_target = param
          if name not in contains and not optional:
            raise Exception("Context %s has required %s import '%s' from context %s, but this context doesn't have that %s" % (orig_source.contextname, sp, name, orig_target.contextname, sp))

      finished = False
      first = True
      while not finished:
        first = False
        finished = True
        for action, param in todo:
          if action == "alias":
            name, newname, optional, orig_source, orig_target = param
            if name not in sockplug:
              finished = False
              continue
            if newname not in sockplug: sockplug[newname] = []
            for item in sockplug[name]:
              if item not in sockplug[newname]:
                finished = False
                sockplug[newname].append(item)
            aliases[name] = newname
          else:
            if not first: continue
            contextinstance, name, newname, optional, orig = param
            retrieved = retrieve(contextinstance, name) 
            if retrieved is None and optional == False:
              raise Exception("Context %s has required %s import '%s' from context %s, but this context doesn't have that %s" % (orig.contextname, sp, name, contextinstance.contextname, sp))
            for item in retrieved:
              if item not in sockplug[newname]:
                sockplug[newname].append(item)
              
    c = self.contexts[-1]
    for k in self.plugins.keys():
      if k not in c.plugins: c.plugins[k] = []
      for p in self.plugins[k]:
        if p not in c.plugins[k]: c.plugins[k].append(p)
    for k in self.sockets.keys():
      if k not in c.sockets: c.sockets[k] = []
      for s in self.sockets[k]:
        if s not in c.sockets[k]: c.sockets[k].append(s)
    self.plugins = {}
    self.sockets = {}
    self.socket_imports = []
    self.plugin_imports = []
    self.socket_imps = set()
    self.plugin_imps = set()
    self.resolved = True


def detect_network(context,extrasock=None,extraplug=None):
  from .context import find_subcontexts
  subcontexts = find_subcontexts(context)
  ret = []
  if extrasock is None and extraplug is None:
    first = True
    extrasock, extraplug = [], []
  else: 
    first = False
  for s in subcontexts:
    if s.contextname[-1].startswith("<<"): continue
    if s.contextname[-1] in ("evin", "evout", "everr", "evexc"): continue
    ok = True
    if not s.import_all_from_parent: 
      ok = False
    else:
      skip = s.import_parent_skip
      for sk in skip:
        #print("SK", sk, bee_evexc(sk))
        if not bee_evexc(sk): continue
        ok = False
        break
    if ok:
      extrasock0, extraplug0 = [], []    
      for sock in s.import_parent_sockets:
        #print(sock)
        if not bee_evexc(sock): extrasock0.append(sock)
      for plug in s.import_parent_plugins:
        #print(plug)
        if not bee_evexc(plug): extraplug0.append(plug)          
      extrasock0.sort()
      extraplug0.sort()
      if first:
        extrasock = extrasock0
        extraplug = extraplug0
        first = False
      else:
       if extrasock0 != extrasock: ok = False
       if extraplug0 != extraplug: ok = False        
    
    if ok:
      newcontexts, dum1, dum2 = detect_network(s,extrasock,extraplug)
      ret += newcontexts      
    else:
      network(s)
  ret.append(context)
  _detect_networks.add(context)
  return ret, extrasock, extraplug

def network(context):
  from . import use_networks
  if not use_networks: return
  if context in _detect_networks: return
  _detect_networks.add(context)
  nw, extrasock, extraplug = detect_network(context) 
  if len(nw) > 1:
    #print("NETWORK", [".".join(c.contextname) for c in nw])
    net = Network(nw)
    for c in nw:
      networks[c] = net
      for k in list(c.plugins.keys()):
        if bee_evexc(k): 
          if k not in extraplug: continue
        if k not in net.plugins: net.plugins[k] = []
        net.plugins[k] += c.plugins[k]
        #print("PLUGINS", c.contextname, k, len(c.plugins[k]))
        del c.plugins[k]
      for k in list(c.sockets.keys()):
        if bee_evexc(k): 
          if k not in extrasock: continue
        if k not in net.sockets: net.sockets[k] = []
        net.sockets[k] += c.sockets[k]
        #print("SOCKETS", c.contextname, k, len(c.sockets[k]))        
        del c.sockets[k]
      for at in "socket_imports", "plugin_imports":
        sp = []
        for contextinstance, name, newname, optional in getattr(c,at):
          if contextinstance in nw:
            if name != newname:
              getattr(net, at).append((nw[-1],name,newname,optional, c))
          else: 
            sp.append((contextinstance, name, newname, optional))
        setattr(c,at,sp)
      
